package com.bocloud.cmp.driver;

import com.bocloud.cmp.model.Butler;
import com.bocloud.cmp.provider.SuperProvider;
import com.bocloud.cmp.provider.common.AuthProvider;
import com.bocloud.common.annotations.Provider;
import com.bocloud.common.model.Result;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Repository;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * @Description：云平台仓库，用来提供相关接口的实现类
 * @Author ：丁明威
 * @Date ：Created in 10:53 2018/11/13
 * @Modified By：
 */
@Repository("providerRepository")
public class ProviderRepository {

    private static Logger logger = LoggerFactory.getLogger(ProviderRepository.class);
    // 缓存provider映射类，key=${code}:${class} value=${specific class full name}
    private static Map<String, Class<?>> providerMap = new ConcurrentHashMap<>();
    private static String PREFIX = "bocloud.cmp.";

    /**
     * 获取指定编码的云平台操作实例，例如，获取VMware平台的HostProvider的实例
     *
     * @param code   云平台编码 如：bocloud.cmp.vmware,bocloud.cmp.openstack等
     * @param clazz  资源操作接口类 如：HostProvider
     * @param butler 云平台认证信息
     * @return 指定编码的资源操作接口实现类，如VMwareHostProvider的实例
     */
    public Object getProvider(String code, Class<?> clazz, Butler butler) {
        try {
            Class<?> providerClass = getProviderClass(code, clazz);
            Constructor<?> constructor = providerClass.getDeclaredConstructor(Butler.class);
            return constructor.newInstance(butler);
        } catch (NoSuchMethodException e) {
            logger.error("init provider {} of {} NoSuchMethodException!", clazz, code, e);
        } catch (InstantiationException e) {
            logger.error("init provider {} of {} InstantiationException!", clazz, code, e);
        } catch (InvocationTargetException e) {
            logger.error("init provider {} of {} InvocationTargetException!", clazz, code, e);
        } catch (IllegalAccessException e) {
            logger.error("init provider {} of {} IllegalAccessException!", clazz, code, e);
        } catch (Exception e) {
            logger.error("get specific provider class {} of {} exception", clazz, code, e);
        }
        return null;
    }

    public Result accessable(Butler butler) {
        if (null == butler || StringUtils.isEmpty(butler.getCategory())) {
            return Result.FAILED("butler is empty!");
        }
        String code = GenericDriver.getProviderValueByButler(butler);
        try {
            // 2020-12-10 这里原来是SuperProvider，现在改成AuthProvider。
            // 理由：应该只有AuthProvider再各个Provider都有实现，而classList.get(0)就不确定拿到哪个Provider了
            // 当然，如果有某个Provider没有实现AuthProvider，那就。。。难了。。。
            List<Class<?>> classList = getProviderClasses(code, AuthProvider.class);
            if (classList.isEmpty()) {
                logger.error("no child class of {} found!", AuthProvider.class);
                return Result.FAILED("平台认证失败!");
            }
            Class<?> clazz = classList.get(0);
            Constructor<?> constructor = clazz.getDeclaredConstructor(Butler.class);
            Object object = constructor.newInstance(butler);
            SuperProvider provider = (SuperProvider) object;
            if (provider.accessable()) {
                return Result.SUCCESS("success");
            }
            return Result.FAILED("平台认证失败！");
        } catch (NoSuchMethodException e) {
            logger.error("NoSuchMethodException！", e);
        } catch (IllegalAccessException e) {
            logger.error("IllegalAccessException！", e);
        } catch (InstantiationException e) {
            logger.error("InstantiationException！", e);
        } catch (InvocationTargetException e) {
            if (e.getTargetException() != null) {
                logger.error("InvocationTargetException！", e.getTargetException());
            } else {
                logger.error("InvocationTargetException！", e);
            }

        } catch (Exception e) {
            logger.error("get specific provider class {} of {} exception", SuperProvider.class, code, e);
        }
        return Result.FAILED("平台认证异常！");
    }

    private List<Class<?>> getProviderClasses(String code, Class<?> clazz) {
        String pkName = clazz.getPackage().getName();
        List<Class<?>> allClass = getAllClass(pkName);
        String extName;
        if (code.contains(".v63")) {
            String commName = pkName.substring(0, pkName.lastIndexOf("."));
            extName = pkName.replace(commName, "com.bocloud.cmp.provider.v63");
            allClass.addAll(getAllClass(extName));
        } else if (code.contains(".v68")) {
            String commName = pkName.substring(0, pkName.lastIndexOf("."));
            extName = pkName.replace(commName, "com.bocloud.cmp.provider.v68");
            allClass.addAll(getAllClass(extName));
        }
        List<Class<?>> fixClass = new ArrayList<>();
        for (Class<?> providerClass : allClass) {
            if (clazz.isAssignableFrom(providerClass)) {
                if (clazz.equals(providerClass)) {
                    continue;
                }
                if (providerClass.isAnnotationPresent(Provider.class)) {
                    Provider provider = (Provider) providerClass.getAnnotation(Provider.class);
                    if (provider.value().equalsIgnoreCase(code)) {
                        fixClass.add(providerClass);
                    }
                }
            }
        }
        return fixClass;
    }

    public synchronized Class<?> getProviderClass(String code, Class<?> clazz) throws Exception {
        if (providerMap.containsKey(buildKey(code, clazz))) {
            return providerMap.get(buildKey(code, clazz));
        }
        List<Class<?>> fixClass = getProviderClasses(code, clazz);
        if (fixClass.isEmpty()) {
            throw new Exception("class not found");
        }
        if (fixClass.size() > 1) {
            throw new Exception("more than one class found!");
        }
        providerMap.put(buildKey(code, clazz), fixClass.get(0));
        return fixClass.get(0);
    }

    private List<Class<?>> getAllClass(String packageName) {
        List<Class<?>> list = new ArrayList<>();
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        String path = packageName.replace(".", "/");
        try {
            Enumeration<URL> enumeration = classLoader.getResources(path);
            while (enumeration.hasMoreElements()) {
                URL url = enumeration.nextElement();
                if ("jar".equals(url.getProtocol())) {
                    String jarPath = url.getPath();
                    if (!jarPath.contains("provider")) {
                        continue;
                    }
                    JarFile jarFile = ((JarURLConnection) url.openConnection()).getJarFile();
                    list.addAll(findClassFromJar(jarFile, path));
                } else {
                    list.addAll(findClassFromFile(new File(url.getFile()), packageName));
                }
            }
        } catch (IOException e) {
            logger.error("get all class of package {} IOException!", packageName, e);
        }
        return list;
    }

    public static List<Class<?>> findClassFromJar(JarFile jarFile, String path) {
        List<Class<?>> classList = new ArrayList<>();
        String packageName = path.replace(".", "/");
        try {
            Enumeration<JarEntry> entries = jarFile.entries();
            String entryName = null;
            while (entries.hasMoreElements()) {
                JarEntry entry = entries.nextElement();
                if (entry.isDirectory()) {
                    continue;
                }
                entryName = entry.getName();
                if (entryName.charAt(0) == '/') {
                    entryName = entryName.substring(1);
                }
                if (!entryName.startsWith(packageName) || !entryName.endsWith(".class")) {
                    continue;
                }
                String className = entryName.substring(0, entryName.length() - 6);
                Class<?> clazz = loadClass(className.replace("/", "."));
                if (null != clazz) {
                    classList.add(clazz);
                }
            }
        } catch (Exception e) {
            logger.error("get class from jar {} error!", jarFile.getName(), e);
        } finally {
            if (null != jarFile) {
                try {
                    jarFile.close();
                } catch (Exception e) {
                    logger.error("close jar {} error!", jarFile.getName(), e);
                }
            }
        }
        return classList;
    }

    private List<Class<?>> findClassFromFile(File file, String packageName) {
        List<Class<?>> list = new ArrayList<>();
        if (!file.exists()) {
            return list;
        }
        File[] files = file.listFiles();
        for (File classFile : files) {
            if (classFile.isDirectory()) {
                if (!classFile.getName().contains(".")) {
                    List<Class<?>> arrayList = findClassFromFile(classFile, packageName + "." + classFile.getName());
                    list.addAll(arrayList);
                }
            } else if (classFile.getName().endsWith(".class")) {
                try {
                    // 保存的类文件不需要后缀.class
                    list.add(Class.forName(
                            packageName + '.' + classFile.getName().substring(0, classFile.getName().length() - 6)));
                } catch (ClassNotFoundException e) {
                    logger.error("get class of {} occur ClassNotFoundException!", file, e);
                }
            }
        }
        return list;
    }

    private static String buildKey(String code, Class<?> clazz) {
        return code + ":" + clazz.getName();
    }

    private static Class<?> loadClass(String fullClassName) {
        try {
            return Thread.currentThread().getContextClassLoader().loadClass(fullClassName);
        } catch (ClassNotFoundException e) {
            logger.error("load class of {} occur ClassNotFoundException!", e);
        }
        return null;
    }
}